
package de.caff.asteroid;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;

/**
 *
 *  A game object has a position and a size and it can be drawn.
 */
public abstract class SizedGameObject
        extends GameObject
{
  /** TEST: calculate maximum velocities? */
  protected static final boolean WATCH_VELOCITIES = false;

  // if you are interested which y coordinates are used change the "if (false)" line in contructor to "if (true)"
  private static int minY = Integer.MAX_VALUE;
  private static int maxY = Integer.MIN_VALUE;

  /** Size of head of velocity arrow. */
  private static final int ARROW_HEAD_SIZE = 8;

  /** The arrow head for velocity drawings. */
  private static final GeneralPath ARROW_HEAD = new GeneralPath();
  static {
    ARROW_HEAD.moveTo(0, 0);
    ARROW_HEAD.lineTo(-ARROW_HEAD_SIZE, ARROW_HEAD_SIZE);
    ARROW_HEAD.lineTo(-ARROW_HEAD_SIZE, -ARROW_HEAD_SIZE);
    ARROW_HEAD.closePath();
  }

  /** The x coordinate of the velocity vector. */
  private int vx;
  /** The y coordinate of the velocity vector. */
  private int vy;

  /**
   *  Constructor.
   *  @param x  x coordinate
   *  @param y  y coordinate
   */
  protected SizedGameObject(int x, int y)
  {
    super(x, y);
    if (false) {
      if (y < minY) {
        minY = y;
        System.out.println("new min/max: "+minY+","+maxY);
      }
      if (y > maxY) {
        maxY = y;
        System.out.println("new min/max: "+minY+","+maxY);
      }
    }
  }

  /**
   *  Get the center of the object.
   *  @return center point
   */
  public Point getCenter()
  {
    return new Point(x, y);
  }

  /**
   *  Get the size of the object.
   *
   *  The size returned by this method is half the length of a square which contains the object,
   *  so the object's bounding box is between (x - size, y - size) and (x + size, y + size).
   *  @return object size
   */
  public abstract int getSize();

  /**
   *  Draw the object.
   *  @param g graphics context
   */
  public void draw(Graphics2D g)
  {
    g.setColor(Color.white);
    int size = getSize();
    g.drawOval(x - size, y - size, 2*size, 2*size);

    drawVelocityVector(g, Color.red);
  }

  /**
   *  Get the shortest vector from this to another game object.
   *
   *  Because of the torus topology it may be shorter going across the borders.
   *  @param obj other object
   *  @return shortest vector from this to the other object
   */
  public Point getDelta(SizedGameObject obj)
  {
    int dx = obj.x - x;
    if (dx < -EXTENT_X/2) {
      dx += EXTENT_X;
    }
    else if (dx >= EXTENT_X/2) {
      dx -= EXTENT_X;
    }
    int dy = obj.y - y;
    if (dy < -EXTENT_Y/2) {
      dy += EXTENT_Y;
    }
    else if (dy >= EXTENT_Y/2) {
      dy -= EXTENT_Y;
    }
    return new Point(dx, dy);
  }

  /**
   *  Get the squared size of this object.
   *  @return squared size
   */
  public int getSquaredSize()
  {
    int size = getSize();
    return size*size;
  }

  /**
   *  Set the velocity.
   *
   *  The velocity is the step between frames.
   *
   *  It is not calculated internally but has to be set from outside. For an example see
   *  {@link de.caff.asteroid.SimpleVelocityPreparer#prepareFrames(java.util.LinkedList)}.
   *  @param x x coordinate of velocity
   *  @param y y coordinate of velocity
   */
  public void setVelocity(int x, int y)
  {
    vx = x;
    vy = y;
  }

  /**
   *  Set the velocity.
   *
   *  The velocity is the step between frames.
   *  @param v velocity vector
   */
  public void setVelocity(Point v)
  {
    setVelocity(v.x, v.y);
  }

  /**
   *  Set the velocity assuming that the given object is at the same place as
   *  this object in the last frame.
   *
   *  @param obj comparision object (<code>null</code> allowed, but then nothing happens)
   */
  public void setVelocityFromDelta(SizedGameObject obj)
  {
    if (obj != null) {
      setVelocity(obj.getDelta(this));
    }
  }

  /**
   *  Get the velocity in x.
   *  @return x component of velocity vector
   */
  public int getVelocityX()
  {
    return vx;
  }

  /**
   *  Get the velocity in y.
   *  @return y component of velocity vector
   */
  public int getVelocityY()
  {
    return vy;
  }

  /**
   *  Get the velocity vector.
   *  @return velocity vector (movement between frames)
   */
  public Point getVelocity()
  {
    return new Point(vx, vy);
  }

  /**
   *  Get the velocity angle.
   *
   *  The angle is measured counterclockwise, with <code>0</code> pointing to the right
   *  @return velocity in radians, or -Math.PI if ship has no velocity
   */
  public double getVelocityAngle()
  {
    return vx != 0 || vy != 0  ?  Math.atan2(vy, vx)  :  -Math.PI;
  }

  /**
   *  Has this object a velocity.
   *  @return velocity
   */
  public boolean hasKnownVelocity()
  {
    return vx != 0 || vy != 0;
  }

  /**
   *  Draw a vector displaying the current velocity.
   *  @param g graphics context
   *  @param color color to use
   */
  protected void drawVelocityVector(Graphics2D g, Color color)
  {
    if (vx != 0  ||  vy != 0) {
      int scale = 16;
      g.setColor(color);
      g.drawLine(x, y, x + scale*vx, y + scale*vy);
      double angle = Math.atan2(vy, vx);
      AffineTransform at = AffineTransform.getTranslateInstance(x + scale*vx, y + scale*vy);
      at.concatenate(AffineTransform.getRotateInstance(angle));
      g.fill(at.createTransformedShape(ARROW_HEAD));
    }
  }
  
  public Point getFuturePosition(int frameStep) {
	  Point futurePos = new Point();
	  
	  futurePos.x = x + (getVelocityX() * frameStep);
	  futurePos.y = y + (getVelocityY() * frameStep);
	  
	  return futurePos;
  }
}
